
/* interp.c -- interpolation / resample functions */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h"
#include "interp.h"

static	float	*tables;

/* set up pre-calculated tables for integer resample ratio	*/

void	initialize_window_tables (void)
{
	int	nr_tables, i2, k;
	double	x0, x1, x2, d, w;
	float	*ft;

	/* x0: phase increment */
	x0 = resample_ratio; if (x0 > 1.0) x0 = 1.0 / x0;
	nr_tables = (int) ((1.0 / x0) + 0.5) >> 1;
	if (!(nr_tables)) {		/* resample ratio = 1	*/
		tables = NULL; return;	/* no tables		*/
	}

	/* allocate memory for tables */

	tables = (float *) malloc (nr_tables * interp_size * sizeof (float));
	if (tables == NULL) {
		fprintf (stderr, "snd_resample: not enough memory\n");
		exit (-1);
	}

	ft = tables; x1 = 0.0; k = nr_tables;
	while (k--) {
		x1 += x0;
		i2 = (interp_size >> 1);
		x2 = sin (x1 * PI) / PI;
		if (resample_ratio < 1.0) x2 *= resample_ratio;
		d = (double) (1 - i2) - x1;

		while (i2--) {
			/* window function */
			w = 1.0 - d * d * win_fact; w *= (w / d++);
			*(ft++) = x2 * w;
			w = 1.0 - d * d * win_fact; w *= (w / d++);
			*(ft++) = -x2 * w;
		}
	}
}

/* free memory used by window tables */

void	free_window_tables (void)
{
	if (tables == NULL) return;
	free (tables); tables = NULL;
}

/* upsample function						*/
/*   in_pos:	input buffer position in samples		*/
/*   out_pos:	output buffer position in samples (integer)	*/

void	upsample (double in_pos, int out_pos)
{
	int	i2, xpos;
	double	x1, x2, d, d2, d2x, d2y, w, n0, n1, n2, n3, *ibuf, *obuf;

	x1 = in_pos; xpos = (int) x1; x1 -= (double) xpos;
	obuf = out_buf + (out_pos * nchnls);

	/* integer sample position: do not interpolate	*/

	if ((x1 < 0.00000001) || (x1 > 0.99999999)) {
		if (x1 > 0.5) xpos++;
		ibuf = in_buf + (xpos * nchnls);
		i2 = nchnls; while (i2--) *(obuf++) = *(ibuf++);
		return;
	}

	n0 = n1 = n2 = n3 = 0.0;
	i2 = (interp_size >> 1);
	x2 = sin (x1 * PI) / PI;
	xpos += (1 - i2);
	d = (double) (1 - i2) - x1;
	d2 = 1.0 - d * d * win_fact;
	d2x = -(2.0 * d * win_fact + win_fact);
	d2y = -2.0 * win_fact;
	ibuf = in_buf + (xpos * nchnls);

	switch (nchnls) {
		case 4:	while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 += *(ibuf++) * w; n1 += *(ibuf++) * w;
				n2 += *(ibuf++) * w; n3 += *(ibuf++) * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 -= *(ibuf++) * w; n1 -= *(ibuf++) * w;
				n2 -= *(ibuf++) * w; n3 -= *(ibuf++) * w;
			}
			*(obuf++) = x2 * n0; *(obuf++) = x2 * n1;
			*(obuf++) = x2 * n2; *(obuf++) = x2 * n3; break;
		case 2:	while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 += *(ibuf++) * w; n1 += *(ibuf++) * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 -= *(ibuf++) * w; n1 -= *(ibuf++) * w;
			}
			*(obuf++) = x2 * n0; *(obuf++) = x2 * n1; break;
		default: while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 += *(ibuf++) * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				n0 -= *(ibuf++) * w;
			}
			*(obuf++) = x2 * n0;
	}
}

/* downsample function						*/
/*   in_pos:	input buffer position in samples (integer)	*/
/*   out_pos:	output buffer position in samples		*/

void	downsample (int in_pos, double out_pos)
{
	int	i2, xpos;
	double	x1, x2, d, d2, d2x, d2y, w, n0, n1, n2, n3, *ibuf, *obuf;

	x1 = out_pos; xpos = (int) x1; x1 -= (double) xpos;
	ibuf = in_buf + (in_pos * nchnls);

	/* integer sample position: do not interpolate	*/

	if ((x1 < 0.00000001) || (x1 > 0.99999999)) {
		if (x1 > 0.5) xpos++;
		obuf = out_buf + (xpos * nchnls);
		i2 = nchnls; while (i2--)
			*(obuf++) += *(ibuf++) * resample_ratio;
		return;
	}

	i2 = (interp_size >> 1);
	x2 = resample_ratio * sin (x1 * PI) / PI;
	xpos += (1 - i2);
	d = (double) (1 - i2) - x1;
	d2 = 1.0 - d * d * win_fact;
	d2x = -(2.0 * d * win_fact + win_fact);
	d2y = -2.0 * win_fact;
	obuf = out_buf + (xpos * nchnls);

	switch (nchnls) {
		case 4:	n0 = *(ibuf++) * x2; n1 = *(ibuf++) * x2;
			n2 = *(ibuf++) * x2; n3 = *(ibuf++) * x2;
			while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) += n0 * w; *(obuf++) += n1 * w;
				*(obuf++) += n2 * w; *(obuf++) += n3 * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) -= n0 * w; *(obuf++) -= n1 * w;
				*(obuf++) -= n2 * w; *(obuf++) -= n3 * w;
			}
			break;
		case 2:	n0 = *(ibuf++) * x2; n1 = *(ibuf++) * x2;
			while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) += n0 * w; *(obuf++) += n1 * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) -= n0 * w; *(obuf++) -= n1 * w;
			}
			break;
		default: n0 = *(ibuf++) * x2;
			while (i2--) {
				/* window function */
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) += n0 * w;
				w = d2 * d2 / d; d++; d2 += d2x; d2x += d2y;
				*(obuf++) -= n0 * w;
			}
	}
}

/* upsample function (integer resample ratio)			*/
/*   in_pos:	input buffer position in samples		*/
/*   out_pos:	output buffer position in samples (integer)	*/

void	upsample_int (double in_pos, int out_pos)
{
	int	i, xpos, ftnum, ft_incr;
	double	x1, w, n0, n1, n2, n3, *ibuf, *obuf;
	float	*ft;

	obuf = out_buf + (out_pos * nchnls);
	x1 = in_pos; xpos = (int) x1; x1 -= (double) xpos;
	if (x1 > 0.5) {
		ft_incr = -1; x1 = 1.0 - x1;
	} else {
		ft_incr = 1;
	}
	ftnum = (int) ((x1 * resample_ratio) + 0.5) - 1;

	/* integer sample position: do not interpolate	*/

	if (ftnum < 0) {
		if (ft_incr < 0) xpos++;
		ibuf = in_buf + (xpos * nchnls);
		i = nchnls; while (i--) *(obuf++) = *(ibuf++);
		return;
	}

	n0 = n1 = n2 = n3 = 0.0;
	ft = tables + (interp_size * ftnum);
	if (ft_incr < 0) ft += interp_size + ft_incr;
	i = interp_size;
	xpos += (1 - (i >> 1));
	ibuf = in_buf + (xpos * nchnls);

	switch (nchnls) {
		case 4:	while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				n0 += *(ibuf++) * w; n1 += *(ibuf++) * w;
				n2 += *(ibuf++) * w; n3 += *(ibuf++) * w;
			}
			*(obuf++) = n0; *(obuf++) = n1;
			*(obuf++) = n2; *(obuf++) = n3; break;
		case 2:	while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				n0 += *(ibuf++) * w; n1 += *(ibuf++) * w;
			}
			*(obuf++) = n0; *(obuf++) = n1; break;
		default: while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				n0 += *(ibuf++) * w;
			}
			*(obuf++) = n0;
	}
}

/* downsample function (integer resample ratio)			*/
/*   in_pos:	input buffer position in samples (integer)	*/
/*   out_pos:	output buffer position in samples		*/

void	downsample_int (int in_pos, double out_pos)
{
	int	i, xpos, ftnum, ft_incr;
	double	x1, w, n0, n1, n2, n3, *ibuf, *obuf;
	float	*ft;

	ibuf = in_buf + (in_pos * nchnls);
	x1 = out_pos; xpos = (int) x1; x1 -= (double) xpos;
	if (x1 > 0.5) {
		ft_incr = -1; x1 = 1.0 - x1;
	} else {
		ft_incr = 1;
	}
	ftnum = (int) ((x1 / resample_ratio) + 0.5) - 1;

	/* integer sample position: do not interpolate	*/

	if (ftnum < 0) {
		if (ft_incr < 0) xpos++;
		obuf = out_buf + (xpos * nchnls);
		i = nchnls; while (i--)
			*(obuf++) += *(ibuf++) * resample_ratio;
		return;
	}

	ft = tables + (interp_size * ftnum);
	if (ft_incr < 0) ft += interp_size + ft_incr;
	i = interp_size;
	xpos += (1 - (i >> 1));
	obuf = out_buf + (xpos * nchnls);

	switch (nchnls) {
		case 4:	n0 = *(ibuf++); n1 = *(ibuf++);
			n2 = *(ibuf++); n3 = *(ibuf++);
			while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				*(obuf++) += n0 * w; *(obuf++) += n1 * w;
				*(obuf++) += n2 * w; *(obuf++) += n3 * w;
			}
			break;
		case 2:	n0 = *(ibuf++); n1 = *(ibuf++);
			while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				*(obuf++) += n0 * w; *(obuf++) += n1 * w;
			}
			break;
		default: n0 = *(ibuf++);
			while (i--) {
				/* window function */
				w = (double) *ft; ft += ft_incr;
				*(obuf++) += n0 * w;
			}
			break;
	}
}

